/***********************************************************************

  File:   cxxabi.h
  Author: Jeffrey Oldham
  Date:   08/15/2000

  Contents:
    RTTI Declarations.

  Copyright (c) 2000 by CodeSourcery, LLC.  All rights reserved.

***********************************************************************/

#ifndef _CSL_CXXABI_H
#define _CSL_CXXABI_H

/***********************************************************************
  Notes
***********************************************************************/

/***********************************************************************
  Included Files
***********************************************************************/

#include <typeinfo>
#include "_dll.hpp"
#include "_utilities.hpp"

/***********************************************************************
  Macros
***********************************************************************/

namespace __cxxabiv1 {

/***********************************************************************
  Forward Declarations
***********************************************************************/

/***********************************************************************
  Type Declarations
***********************************************************************/

typedef _DoublyLinkedListDeallocate<_ObjectPathPair> _object_list_t;

/***********************************************************************
  Class Declarations
***********************************************************************/

class __fundamental_type_info
  : public std::type_info
{
  virtual 	~__fundamental_type_info ();
};

class __array_type_info
  : public std::type_info
{
  virtual 	~__array_type_info ();
};

class __function_type_info
  : public std::type_info
{
  virtual 	~__function_type_info ();
};

class __enum_type_info
  : public std::type_info
{
  virtual	~__enum_type_info ();
};

/* Used for class types having no bases and is also a base type for
   the other two class type representations.  */

class __class_type_info
  : public std::type_info
{

public:

  virtual void	_FindMatchingBases (_const_object_t
				    _object,
				    const bool
				    _is_reached_via_public_path,
				    const __class_type_info&
				    _desired_type,
				    _object_list_t&
				    _objects_with_desired_type)
    const;
			/* Preconditions:
			     _OBJECT is the non-zero object
			     corresponding to *THIS type information.
			     _IS_REACHED_VIA_PUBLIC_PATH is true iff
			     this object has been reach only through
			     public derivations.  _DESIRED_TYPE is a
			     non-zero pointer to a structure
			     encapsulating the desired type.
			   Postconditions:
			     All bases, including possibly this one,
			     that have type matching _DESIRED_TYPE, are
			     added to _OBJECTS_WITH_DESIRED_TYPE.  */

  virtual bool	_FindMatchingBaseObject (_const_object_t _object,
					 _const_object_t _base_object,
					 const bool	 _check_for_public)
    const;
			/* Preconditions:
			     _OBJECT is a non-zero object.
			     _BASE_OBJECT is a non_zero object.
			   Postconditions:
			     Returns true iff _BASE_OBJECT is an
			     accessible base object of _OBJECT.  If
			     _CHECK_FOR_PUBLIC is true, only publicly
			     accessible paths are used.  */
  
  virtual bool	_eh_match (const type_info&         _caught_type,
			   _const_object_t          _object,
			   _const_object_t*         _adjusted_object,
			   const bool		    _const_so_far,
			   const unsigned	    _pointer_level) const;
			/* For documentation, see
			   type_info::_eh_match ().  */

};

/* Used for classes containing only a single, public, non-virtual
   base at offset zero, i.e., the derived class is dynamic iff the
   base is.  */

class __si_class_type_info
  : public __class_type_info
{

public:

  const __class_type_info*
		__base_type;

  virtual void	_FindMatchingBases (_const_object_t
				    _object,
				    const bool
				    _is_reached_via_public_path,
				    const __class_type_info&
				    _desired_type,
				    _object_list_t&
				    _objects_with_desired_type)
    const;
			/* See __class_type_info::_FindMatchingBases.  */

  virtual bool	_FindMatchingBaseObject (_const_object_t _object,
					 _const_object_t _base_object,
					 const bool	 _check_for_public)
    const;
			/* See __class_type_info::_FindMatchingBaseObject.  */

  virtual bool	_eh_match (const type_info&         _caught_type,
			   _const_object_t          _object,
			   _const_object_t*         _adjusted_object,
			   const bool		    _const_so_far,
			   const unsigned	    _pointer_level) const;
			/* For documentation, see
			   type_info::_eh_match ().  */

};

/* Used for base class descriptions.  */
class __base_class_type_info
{
public:

  const __class_type_info*
		__base_type;
			/* The __base_type member points to the RTTI
                           for the base type.  */

  long		__offset_flags;
			/* The upper 56 bits of __offset_flags are a
			   signed offset. For a non-virtual base, this
			   is the offset in the object of the base
			   subobject. For a virtual base, this is the
			   offset in the virtual table of the virtual
			   base offset for the virtual base referenced
			   (negative).  */

  enum		__offset_flags_masks {

    		__virtual_mask = 0x1,
			/* Base class is virtual.  */

		__public_mask = 0x2,
			/* Base class is public.  */

		__offset_shift = 8
  };

  bool		_IsPublicBaseClass () const;
			/* Returns:
			     True iff the base class is a public base
			     class.  */

  _object_t	_ObtainBaseObject (_const_object_t _derived_object) const;
			/* Preconditions:
			     _DERIVED_OBJECT is the object for this
			     base's derived class.
			   Returns:
			     The object corresponding to the base.  */

};

/* Used for classes with bases that do not satisfy the
   __si_class_type_info constraints.  */

class __vmi_class_type_info
  : public __class_type_info
{
public:

  unsigned int	__flags;
			/* __flags is a word with flags describing
			   details about the class structure, which
			   may be referenced by using the
			   __flags_masks enumeration.  */

  unsigned int	__base_count;
			/* This is a word with the number of direct
			   proper base class descriptions that follow.  */

  __base_class_type_info
    		__base_info[1];
			/* This is an array of base class descriptions
			   -- one for every direct proper base.  The
			   resulting structure is variable-length,
			   with the actual size depending on the
			   number of trailing base class descriptions.  */

  enum __flags_masks {

		__non_diamond_repeat_mask = 0x1,
			/* Class has non-diamond repeated inheritance.  */

    		__diamond_shaped_mask = 0x2
			/* Class is diamond shaped.  */

  };

  virtual void	_FindMatchingBases (_const_object_t
				    _object,
				    const bool
				    _is_reached_via_public_path,
				    const __class_type_info&
				    _desired_type,
				    _object_list_t&
				    _objects_with_desired_type) const;
			/* See __class_type_info::_FindMatchingBases.  */

  virtual bool	_FindMatchingBaseObject (_const_object_t _object,
					 _const_object_t _base_object,
					 const bool	 _check_for_public)
    const;
			/* See __class_type_info::_FindMatchingBaseObject.  */

  virtual bool	_eh_match (const type_info&         _caught_type,
			   _const_object_t          _object,
			   _const_object_t*         _adjusted_object,
			   const bool		    _const_so_far,
			   const unsigned	    _pointer_level) const;
			/* For documentation, see
			   type_info::_eh_match ().  */

};

/* Used as a base for both pointer types and pointer-to-member
   types.  */

class __pbase_type_info
  : public std::type_info
{
public:

  unsigned int	__flags;
			/* This is a flag word describing the
			   cv-qualification and other attributes of
			   the type pointed to, e.g., "int volatile*"
			   should have the "volatile" bit set in that
			   word.  */

  const std::type_info*
		__pointee;
			/* This is a pointer to the std::type_info
			   derivation for the unqualified type being
			   pointed to.  */

  enum __masks {

    		__const_mask = 0x1,
			/* __pointee type has const qualifier.  */

    		__volatile_mask = 0x2,
			/* __pointee type has volatile qualifier.  */

    		__restrict_mask = 0x4,
			/* __pointee type has restrict qualifier.  */

    		__incomplete_mask = 0x8,
			/* __pointee type is incomplete.  */

    		__incomplete_class_mask = 0x10
			/* class containing __pointee is incomplete
			   (in pointer to member).  */

  };

};

class __pointer_type_info
  : public __pbase_type_info
{

public:

  virtual bool	_eh_match (const type_info&         _caught_type,
			   _const_object_t          _object,
			   _const_object_t*         _adjusted_object,
			   const bool		    _const_so_far,
			   const unsigned	    _pointer_level) const;
			/* For documentation, see
			   type_info::_eh_match ().  */

};

class __pointer_to_member_type_info
  : public __pbase_type_info
{

public:
    
  const __class_type_info*
    		__context;
			/* This is a pointer to an __class_type_info
			   corresponding to the class type containing
			   the member pointed to, e.g., the "A" in
			   "int A::*".  */

  virtual bool	_eh_match (const type_info&         _caught_type,
			   _const_object_t          ,
			   _const_object_t*         ,
			   const bool		    ,
			   const unsigned	    ) const;
			/* For documentation, see
			   type_info::_eh_match ().  */

};

/***********************************************************************
  Variable Declarations
***********************************************************************/

/***********************************************************************
  Function Declarations
***********************************************************************/

/***********************************************************************
  Inline Function Definitions
***********************************************************************/

} // Namespace __cxxabiv1

namespace abi = __cxxabiv1;

#endif // _CSL_CXXABI_H
