/**
*** Copyright (C) 1985-2001 Intel Corporation.  All rights reserved.
***
*** The information and source code contained herein is the exclusive
*** property of Intel Corporation and may not be disclosed, examined
*** or reproduced in whole or in part without explicit written authorization
*** from the company.
***
*** static char cvs_id[] = "$Id: ia64intrin.h,v 1.1 2001/12/11 23:38:51 jkmishra Exp $";
***
**/

#ifndef _IA64INTRIN_H_INCLUDED
#define _IA64INTRIN_H_INCLUDED

/*
 * the m64 type is required for the integer Streaming SIMD Extensions intrinsics
 */
#ifndef _MMINTRIN_H_INCLUDED
#include <mmintrin.h>
#endif

/*
 * Define __int64 type for __lp64__
 */
#if defined(__LP64__)
#ifndef __int64
#define _IA64INTRIN_H_DEFINED__INT64
#define __int64 long
#endif
#endif

/* Integer values for mux1 instruction */
#define _m64_mux1_brcst 0
#define _m64_mux1_mix   8
#define _m64_mux1_shuf  9
#define _m64_mux1_alt  10
#define _m64_mux1_rev  11

/* Itanium Native Intrinsics Declarations */


#ifdef __cplusplus
extern "C" { /* Don't use c++ name mangling */
#endif

/* Lock and Atomic Operation Related Intrinsics */

unsigned __int64 _InterlockedExchange8(volatile unsigned char *Target, unsigned __int64 value);
unsigned __int64 _InterlockedCompareExchange8_rel(volatile unsigned char *Destination, unsigned __int64 Exchange, unsigned __int64 Comparand);
unsigned __int64 _InterlockedCompareExchange8_acq(volatile unsigned char *Destination, unsigned __int64 Exchange, unsigned __int64 Comparand);

unsigned __int64 _InterlockedExchange16(volatile unsigned short *Target, unsigned __int64 value);
unsigned __int64 _InterlockedCompareExchange16_rel(volatile unsigned short *Destination, unsigned __int64 Exchange, unsigned __int64 Comparand);
unsigned __int64 _InterlockedCompareExchange16_acq(volatile unsigned short *Destination, unsigned __int64 Exchange, unsigned __int64 Comparand);

int _InterlockedIncrement(volatile int *addend);
int _InterlockedDecrement(volatile int *addend);
int _InterlockedExchange(volatile int *Target, int value);
unsigned __int64 _InterlockedExchangeU(volatile unsigned int *Target, unsigned __int64 value);
unsigned __int64 _InterlockedCompareExchange_rel(volatile unsigned int *Destination, unsigned __int64 Exchange, unsigned __int64 Comparand);
unsigned __int64 _InterlockedCompareExchange_acq(volatile unsigned int *Destination, unsigned __int64 Exchange, unsigned __int64 Comparand);
int _InterlockedCompareExchange(volatile int *Destination, int Exchange, int Comparand);

void * _InterlockedCompareExchangePointer(void * volatile *Destination, void *Exchange, void *Comparand);
int _InterlockedExchangeAdd(volatile int *addend, int increment);
int _InterlockedAdd(volatile int *addend, int increment);
void _ReleaseSpinLock(volatile int *x);

__int64 _InterlockedIncrement64(volatile __int64 *addend);
__int64 _InterlockedDecrement64(volatile __int64 *addend);
__int64 _InterlockedExchange64(volatile __int64 *Target, __int64 value);
unsigned __int64 _InterlockedExchangeU64(volatile unsigned __int64 *Target, unsigned __int64 value);
unsigned __int64 _InterlockedCompareExchange64_rel(volatile unsigned __int64 *Destination, unsigned __int64 Exchange, unsigned __int64 Comparand);
unsigned __int64 _InterlockedCompareExchange64_acq(volatile unsigned __int64 *Destination, unsigned __int64 Exchange, unsigned __int64 Comparand);
__int64 _InterlockedCompareExchange64(volatile __int64 *Destination, __int64 Exchange, __int64 Comparand);
__int64 _InterlockedExchangeAdd64(volatile __int64 *addend, __int64 increment);
__int64 _InterlockedAdd64(volatile __int64 *addend, __int64 increment);



/* OS Related Itanium Intrinsics  */

/* The names to use for whichReg and whichIndReg below come from
   the include file ia64regs.h */

#include <ia64regs.h>
unsigned __int64 __getReg(const int whichReg);
void __setReg(const int whichReg, unsigned __int64 value);

/* Operations on Indirect registers:  whichIndReg[index] */
unsigned __int64 __getIndReg(const int whichIndReg, __int64 index);
void __setIndReg(const int whichIndReg, __int64 index, unsigned __int64 value);

#if defined(__LP64__)
void *_rdteb(void);
#else
void *__ptr64 _rdteb(void);
#endif
void __isrlz(void);
void __dsrlz(void);

/* Below, "increment" must be literal: -16, -8, -4, -1, 1, 4, 8, 16 */
unsigned __int64 __fetchadd4_acq(unsigned int *addend, const int increment);
unsigned __int64 __fetchadd4_rel(unsigned int *addend, const int increment);
unsigned __int64 __fetchadd8_acq(unsigned __int64 *addend, const int increment);
unsigned __int64 __fetchadd8_rel(unsigned __int64 *addend, const int increment);

void __fwb(void);

/* In the following, whichFloatReg should be an integer from 0-127 */
void __ldfs(const int whichFloatReg, void *src);
void __ldfd(const int whichFloatReg, void *src);
void __ldfe(const int whichFloatReg, void *src);
void __ldf8(const int whichFloatReg, void *src);
void __ldf_fill(const int whichFloatReg, void *src);
void __stfs(void *dst, const int whichFloatReg);
void __stfd(void *dst, const int whichFloatReg);
void __stfe(void *dst, const int whichFloatReg);
void __stf8(void *dst, const int whichFloatReg);
void __stf_spill(void *dst, const int whichFloatReg);

void __mf(void);
void __mfa(void);
void __synci(void);
void __thash(__int64);
void __ttag(__int64);
void __itcd(__int64 pa);
void __itci(__int64 pa);
void __itrd(__int64 whichTransReg, __int64 pa);
void __itri(__int64 whichTransReg, __int64 pa);
void __ptce(__int64 va);
void __ptcl(__int64 va, __int64 pagesz);
void __ptcg(__int64 va, __int64 pagesz);
void __ptcga(__int64 va, __int64 pagesz);
void __ptri(__int64 va, __int64 pagesz);
void __ptrd(__int64 va, __int64 pagesz);
__int64 __tpa(__int64 va);
void __invalat(void);
void __invala (void);  /* alternative spelling of __invalat */
void __invala_gr(const int whichGeneralReg /* 0-127 */ );
void __invala_fr(const int whichFloatReg /* 0-127 */ );
void __break(const int);
void __nop(const int);
void __debugbreak(void);
void __fc(__int64);
void __sum(int mask);
void __rum(int mask);
void __ssm(int mask);
void __rsm(int mask);

__int64 _ReturnAddress(void);

/* Values for lfhint in __lfetch and __lfetch_fault */

#define	__lfhint_none	0
#define	__lfhint_nt1	1
#define	__lfhint_nt2	2
#define	__lfhint_nta	3

void __lfetch(int lfhint, void *y);
void __lfetch_fault(int lfhint, void *y);




/* Type Conversion */

__int64 _m_to_int64(__m64 a);
__m64 _m_from_int64(__int64 a);
__int64 __round_double_to_int64(double d);
unsigned __int64 __getf_exp(double d);


/* Native Integer Operations */

__int64 _m64_dep_mr(__int64 r, __int64 s, const int pos, const int len); 
__int64 _m64_dep_mi(const int v, __int64 s, const int p, const int len); 
__int64 _m64_dep_zr(__int64 s, const int pos, const int len); 
__int64 _m64_dep_zi(const int v, const int pos, const int len);
__int64 _m64_extr(__int64 r, const int pos, const int len); 
__int64 _m64_extru(__int64 r, const int pos, const int len);
__int64 _m64_xmal(__int64 a, __int64 b, __int64 c); 
__int64 _m64_xmalu(__int64 a, __int64 b, __int64 c); 
__int64 _m64_xmah(__int64 a, __int64 b, __int64 c); 
__int64 _m64_xmahu(__int64 a, __int64 b, __int64 c);
__int64 _m64_popcnt(__int64 a);
__int64 _m64_shladd(__int64 a, const int count, __int64 b);
__int64 _m64_shrp(__int64 a, __int64 b, const int count);

/*  Native fpsr operations */

extern void _fsetc(unsigned int amask, unsigned int omask);
extern void _fclrf(void);  


/* Itanium Multimedia Additions */

__int64 _m64_czx1l(__m64 a);
__int64 _m64_czx1r(__m64 a);
__int64 _m64_czx2l(__m64 a) ;
__int64 _m64_czx2r(__m64 a);
__m64 _m64_mix1l(__m64 a, __m64 b); 
__m64 _m64_mix2l(__m64 a, __m64 b); 
__m64 _m64_mix4l(__m64 a, __m64 b); 
__m64 _m64_mix1r(__m64 a, __m64 b); 
__m64 _m64_mix2r(__m64 a, __m64 b); 
__m64 _m64_mix4r(__m64 a, __m64 b);
__m64 _m64_mux1(__m64 a, const int n); 
__m64 _m64_mux2(__m64 a, const int n);
__m64 _m64_padd1uus(__m64 a, __m64 b); 
__m64 _m64_padd2uus(__m64 a, __m64 b);
__m64 _m64_pavg1_nraz(__m64 a, __m64 b); 
__m64 _m64_pavg2_nraz(__m64 a, __m64 b);
__m64 _m64_pavgsub1(__m64 a, __m64 b); 
__m64 _m64_pavgsub2(__m64 a, __m64 b);
__m64 _m64_pmpy2r(__m64 a, __m64 b); 
__m64 _m64_pmpy2l(__m64 a, __m64 b);
__m64 _m64_pmpyshr2(__m64 a, __m64 b, const int count); 
__m64 _m64_pmpyshr2u(__m64 a, __m64 b, const int count);
__m64 _m64_pshladd2(__m64 a, const int count, __m64 b);
__m64 _m64_pshradd2(__m64 a, const int count, __m64 b);
__m64 _m64_psub1uus(__m64 a, __m64 b); 
__m64 _m64_psub2uus(__m64 a, __m64 b);


#ifdef __cplusplus
}
#endif

/*
 * Undefined __int64 type for __lp64__
 */
#if defined(__LP64__)
#ifdef _IA64INTRIN_H_DEFINED__INT64
#undef __int64
#endif
#endif

#endif /* _IA64INTRIN_H_INCLUDED */

