/* static char cvs_id[] = "$Id: stdarg.h,v 1.2 2001/11/02 19:22:54 ubivatur Exp $"; */

/**
*** Copyright (C) 1985-1999 Intel Corporation.  All rights reserved.
***
*** The information and source code contained herein is the exclusive
*** property of Intel Corporation and may not be disclosed, examined
*** or reproduced in whole or in part without explicit written authorization
*** from the company.
***
*** static char sccs_id[] = "@(#)stdarg.h	1.5 02/11/99 11:12:02";
**/

/*
 * stdarg.h
 *
 * Replacement header for standard stdarg.h
 *
 * This replacement is necessary to support the 16 byte
 * aligned parameters.
 *
 * 18 Sept 1997 [kbs]
 */

#ifndef _INC_STDARG
#define _INC_STDARG


#if defined(__ICC) || defined(__ECC) || defined (__ICL)

#define __builtin_alignof(t) __ALIGNOF__(t)
#ifdef __ICL
    #define __SIZE_TYPE__		unsigned int
#endif

#else

#ifdef _M_IA64
    #define __builtin_alignof(t)	(8)
    #define __SIZE_TYPE__		unsigned long
#else
    #define __builtin_alignof(t)	(4)
    #define __SIZE_TYPE__		unsigned int
#endif /* _M_IA64 */

#endif /* defined(__ICC) | defined(__ECC) */



/*
 * The following macro computes the size actually occupied in
 * the parameter block by a parameter of the same type as t.
 * All parameter sizes are really multiples of int size in a
 * parameter block.
 */
#define _PARMSIZE(t) \
    ((sizeof(t) + (sizeof(__SIZE_TYPE__)-1)) & ~(sizeof(__SIZE_TYPE__) -1))

/*
 * The following macro computes the alignment in the parameter block
 * of a parameter of type t.  Any type that doesn't require 16 byte
 * alignment, only gets 4 byte alignment.
 */
#ifdef _M_IA64
#define _PARMALIGN(t) \
    (__builtin_alignof(t) < 16 ? 8 : 16)

#else 
#define _PARMALIGN(t) \
    (__builtin_alignof(t) < 16 ? 4 : 16)

#endif
/*
 * The following macro computes the base address of a parameter of type
 * t given the address of the first byte after the previous parameter.
 */
#define _PARMADDR(fb_addr,t) \
    ((char *)((((__SIZE_TYPE__)fb_addr) + (_PARMALIGN(t)-1)) & ~(_PARMALIGN(t)-1)))

/*
 * If _VA_LIST_DEFINED is defined, then va_list type is defined in 
 * <yvals.h> that was included before <stdarg.h> and we have no need
 * to define it here. If not, then we'll define this type here and 
 * we'll define _VA_LIST_DEFINED to pevent second definition of 
 * va_list type in <yvals.h> for cases, when it's included after <stdarg.h>.
 */
#ifndef _VA_LIST_DEFINED
typedef char * va_list;
#define _VA_LIST_DEFINED
#endif
/* Needed for some GNU applications */
#ifndef __GNUC_VA_LIST
#define __gnuc_va_list va_list
#define __GNUC_VA_LIST
#endif

#define va_start(ap,anchor) \
    ((ap) = (va_list)&anchor + _PARMSIZE(anchor))

#define va_arg(ap,t) \
    (((t *)(ap = (va_list)(_PARMADDR(ap,t) + sizeof(t))))[-1])

#define va_end(ap) \
    ( ap = 0 )
    
#define va_copy(dest, src) \
    ( dest ) = ( src )

#endif	/* _INC_STDARG */
